form {
    margin: 0;
    padding: 0;
}

button {
    border-color: #444;
    width: auto;
    overflow: visible;
}

.modal-dialog-overlay {
    background: black;
    filter: alpha(opacity=75);
}

.modal-dialog-middle {
    position: absolute;
    top: 5%;
    left: 0%;
}

.modal-dialog-inner {
    margin: 0 auto;
    width: 75%;
    border: 2px solid #999;
}

.methanal-tab-view .methanal-tab-label {
    display: inline;
    zoom: 1;
}

.methanal-tab-view .selected-tab-label {
    border-bottom: 1px solid #fff;
    position: relative;
}

.methanal-tab-view .methanal-tab-group {
    position: relative;
    top: 1px;
}

.dependancy-parent {
    background-position: 3px center;
}

.hover-tooltip {
    /* Simulate inline-block. */
    zoom: 1;
    display: inline;
    vertical-align: top;
}

.label-tooltip .hover-tooltip {
    position: relative;
    top: -0.25em;
    margin-right: -100%;
}

@media print {
    .print-target .modal-dialog-middle {
        top: 0;
    }

    .print-target .modal-dialog-inner {
        /* 100% clips on the edges. */
        width: 99%;
        border: none;
    }
}
