@import url(methanal-ie.css);

.methanal-multi-input {
    margin-left: 3px;
}

.methanal-multi-input .methanal-caption {
    width: 8em;
}

.methanal-caption {
    width: 20em;
}

.dependancy-child {
    border-bottom: 1px solid #ccc;
}

.methanal-control-inputs {
    float: none;
}

.methanal-form-group {
    background-image: none;
}

/**
 * Select elements in IE6 are rendered as window-level objects, which means
 * they poke through the overlay. Rendering a huge IFRAME with the "mask"
 * filter seems to hide them.
 */
.modal-dialog-outer iframe,
.modal-dialog-overlay {
    position: absolute;
    width: 100%;
    height: expression(document.documentElement.clientHeight + 'px');
    left: expression((ignoreMe2 = document.documentElement.scrollLeft ? document.documentElement.scrollLeft : document.body.scrollLeft) + 'px');
    top: expression((ignoreMe = document.documentElement.scrollTop ? document.documentElement.scrollTop : document.body.scrollTop) + 'px');
    right: auto;
    bottom: auto;
}

.modal-dialog-outer iframe {
    zoom: 1;
    z-index: -1;
    filter: mask();
}

.modal-dialog-middle {
    margin-top: 5%;
}

.modal-dialog-inner {
    /* Reset. */
    vertical-align: baseline;
}
